// Copyright 2004 Adam Megacz, see the COPYING file for licensing [GPL]
package org.ibex.js;

/** A JavaScript regular expression object */
public class JSRegexp extends JS {
    private boolean global;
    private GnuRegexp.RE re;
    private int lastIndex;
    
    public JSRegexp(Object arg0, Object arg1) throws JSExn {
        if(arg0 instanceof JSRegexp) {
            JSRegexp r = (JSRegexp) arg0;
            this.global = r.global;
            this.re = r.re;
            this.lastIndex = r.lastIndex;
        } else {
            String pattern = (String)arg0;
            String sFlags = null;
            int flags = 0;
            if(arg1 != null) sFlags = (String)arg1;
            if(sFlags == null) sFlags = "";
            for(int i=0;i<sFlags.length();i++) {
                switch(sFlags.charAt(i)) {
                    case 'i': flags |= GnuRegexp.RE.REG_ICASE; break;
                    case 'm': flags |= GnuRegexp.RE.REG_MULTILINE; break;
                    case 'g': global = true; break;
                    default: throw new JSExn("Invalid flag in regexp \"" + sFlags.charAt(i) + "\"");
                }
            }
            re = newRE(pattern,flags);
            put("source", pattern);
            put("global", B(global));
            put("ignoreCase", B(flags & GnuRegexp.RE.REG_ICASE));
            put("multiline", B(flags & GnuRegexp.RE.REG_MULTILINE));
        }
    }

    public Object callMethod(Object method, Object a0, Object a1, Object a2, Object[] rest, int nargs) throws JSExn {
        switch(nargs) {
            case 1: {
                //#switch(method)
                case "exec": {
                    String s = (String)a0;
                    int start = global ? lastIndex : 0;
                    if(start < 0 || start >= s.length()) { lastIndex = 0; return null; }
                    GnuRegexp.REMatch match = re.getMatch(s,start);
                    if(global) lastIndex = match == null ? s.length() : match.getEndIndex();
                    return match == null ? null : matchToExecResult(match,re,s);
                }
                case "test": {
                    String s = (String)a0;
                    if (!global) return B(re.getMatch(s) != null);
                    int start = global ? lastIndex : 0;
                    if(start < 0 || start >= s.length()) { lastIndex = 0; return null; }
                    GnuRegexp.REMatch match = re.getMatch(s,start);
                    lastIndex = match != null ? s.length() : match.getEndIndex();
                    return B(match != null);
                }
                case "toString": return toString(a0);
                case "stringMatch": return stringMatch(a0,a1);
                case "stringSearch": return stringSearch(a0,a1);
                //#end
                break;
            }
            case 2: {
                //#switch(method)
                case "stringReplace": return stringReplace(a0, a1,a2);
                //#end
                break;
            }
        }
        return super.callMethod(method, a0, a1, a2, rest, nargs);
    }
    
    public Object get(Object key) throws JSExn {
        //#switch(key)
        case "exec": return METHOD;
        case "test": return METHOD;
        case "toString": return METHOD;
        case "lastIndex": return N(lastIndex);
        //#end
        return super.get(key);
    }
    
    public void put(Object key, Object value) throws JSExn {
        if(key.equals("lastIndex")) lastIndex = JS.toNumber(value).intValue();
        super.put(key,value);
    }
  
    private static Object matchToExecResult(GnuRegexp.REMatch match, GnuRegexp.RE re, String s) {
        try {
            JS ret = new JS();
            ret.put("index", N(match.getStartIndex()));
            ret.put("input",s);
            int n = re.getNumSubs();
            ret.put("length", N(n+1));
            ret.put("0",match.toString());
            for(int i=1;i<=n;i++) ret.put(Integer.toString(i),match.toString(i));
            return ret;
        } catch (JSExn e) {
            throw new Error("this should never happen");
        }
    }
    
    public String toString() {
        try {
            StringBuffer sb = new StringBuffer();
            sb.append('/');
            sb.append(get("source"));
            sb.append('/');
            if(global) sb.append('g');
            if(Boolean.TRUE.equals(get("ignoreCase"))) sb.append('i');
            if(Boolean.TRUE.equals(get("multiline"))) sb.append('m');
            return sb.toString();
        } catch (JSExn e) {
            throw new Error("this should never happen");
        }
    }
    
    public static Object stringMatch(Object o, Object arg0) throws JSExn {
        String s = o.toString();
        GnuRegexp.RE re;
        JSRegexp regexp = null;
        if(arg0 instanceof JSRegexp) {
            regexp = (JSRegexp) arg0;
            re = regexp.re;
        } else {
            re = newRE(arg0.toString(),0);
        }
        
        if(regexp == null) {
            GnuRegexp.REMatch match = re.getMatch(s);
            return matchToExecResult(match,re,s);
        }
        if(!regexp.global) return regexp.callMethod("exec", s, null, null, null, 1);
        
        JSArray ret = new JSArray();
        GnuRegexp.REMatch[] matches = re.getAllMatches(s);
        for(int i=0;i<matches.length;i++) ret.addElement(matches[i].toString());
        regexp.lastIndex = matches.length > 0 ? matches[matches.length-1].getEndIndex() : s.length();
        return ret;
    }
    
    public static Object stringSearch(Object o, Object arg0) throws JSExn  {
        String s = o.toString();
        GnuRegexp.RE re = arg0 instanceof JSRegexp ? ((JSRegexp)arg0).re : newRE(arg0.toString(),0);
        GnuRegexp.REMatch match = re.getMatch(s);
        return match == null ? N(-1) : N(match.getStartIndex());
    }
    
    public static Object stringReplace(Object o, Object arg0, Object arg1) throws JSExn {
        String s = o.toString();
        GnuRegexp.RE re;
        JSFunction replaceFunc = null;
        String replaceString = null;
        JSRegexp regexp = null;
        if(arg0 instanceof JSRegexp) {
            regexp = (JSRegexp) arg0;
            re = regexp.re;
        } else {
            re = newRE(arg0.toString(),0);
        }
        if(arg1 instanceof JSFunction)
            replaceFunc = (JSFunction) arg1;
        else
            replaceString = JS.toString(arg1.toString());
        GnuRegexp.REMatch[] matches;
        if(regexp != null && regexp.global) {
            matches = re.getAllMatches(s);
            if(regexp != null) {
                if(matches.length > 0)
                    regexp.lastIndex = matches[matches.length-1].getEndIndex();
                else
                    regexp.lastIndex = s.length();
            }
        } else {
            GnuRegexp.REMatch match = re.getMatch(s);
            if(match != null)
                matches = new GnuRegexp.REMatch[]{ match };
            else
                matches = new GnuRegexp.REMatch[0];
        }
        
        StringBuffer sb = new StringBuffer(s.length());
        int pos = 0;
        char[] sa = s.toCharArray();
        for(int i=0;i<matches.length;i++) {
            GnuRegexp.REMatch match = matches[i];
            sb.append(sa,pos,match.getStartIndex()-pos);
            pos = match.getEndIndex();
            if(replaceFunc != null) {
                int n = (regexp == null ? 0 : re.getNumSubs());
                int numArgs = 3 + n;
                Object[] rest = new Object[numArgs - 3];
                Object a0 = match.toString();
                Object a1 = null;
                Object a2 = null;
                for(int j=1;j<=n;j++)
                    switch(j) {
                        case 1: a1 = match.toString(j); break;
                        case 2: a2 = match.toString(j); break;
                        default: rest[j - 3] = match.toString(j); break;
                    }
                switch(numArgs) {
                    case 3:
                        a1 = N(match.getStartIndex());
                        a2 = s;
                        break;
                    case 4:
                        a2 = N(match.getStartIndex());
                        rest[0] = s;
                        break;
                    default:
                        rest[rest.length - 2] = N(match.getStartIndex());
                        rest[rest.length - 1] = s;
                }

                // note: can't perform pausing operations in here
                sb.append((String)replaceFunc.call(a0, a1, a2, rest, numArgs));

            } else {
                sb.append(mySubstitute(match,replaceString,s));
            }
        }
        int end = matches.length == 0 ? 0 : matches[matches.length-1].getEndIndex();
        sb.append(sa,end,sa.length-end);
        return sb.toString();
    }
    
    private static String mySubstitute(GnuRegexp.REMatch match, String s, String source) {
        StringBuffer sb = new StringBuffer();
        int i,n;
        char c,c2;
        for(i=0;i<s.length()-1;i++) {
           c = s.charAt(i);
            if(c != '$') {
                sb.append(c);
                continue;
            }
            i++;
            c = s.charAt(i);
            switch(c) {
                case '0': case '1': case '2': case '3': case '4':
                case '5': case '6': case '7': case '8': case '9':
                    if(i < s.length()-1 && (c2 = s.charAt(i+1)) >= '0' && c2 <= '9') {
                        n = (c - '0') * 10 + (c2 - '0');
                        i++;
                    } else {
                        n = c - '0';
                    }
                    if(n > 0)
                        sb.append(match.toString(n));
                    break;
                case '$':
                    sb.append('$'); break;
                case '&':
                    sb.append(match.toString()); break;
                case '`':
                    sb.append(source.substring(0,match.getStartIndex())); break;
                case '\'':
                    sb.append(source.substring(match.getEndIndex())); break;
                default:
                    sb.append('$');
                    sb.append(c);
            }
        }
        if(i < s.length()) sb.append(s.charAt(i));
        return sb.toString();
    }
                    
    
    public static Object stringSplit(String s, Object arg0, Object arg1, int nargs) {
        int limit = nargs < 2 ? Integer.MAX_VALUE : JS.toInt(arg1);
        if(limit < 0) limit = Integer.MAX_VALUE;
        if(limit == 0) return new JSArray();
        
        GnuRegexp.RE re = null;
        JSRegexp regexp = null;
        String sep = null;
        JSArray ret = new JSArray();
        int p = 0;
        
        if(arg0 instanceof JSRegexp) {
            regexp = (JSRegexp) arg0;
            re = regexp.re;
        } else {
            sep = arg0.toString();
        }
        
        // special case this for speed. additionally, the code below doesn't properly handle
        // zero length strings
        if(sep != null && sep.length()==0) {
            int len = s.length();
            for(int i=0;i<len;i++)
                ret.addElement(s.substring(i,i+1));
            return ret;
        }
        
        OUTER: while(p < s.length()) {
            if(re != null) {
                GnuRegexp.REMatch m = re.getMatch(s,p);
                if(m == null) break OUTER;
                boolean zeroLength = m.getStartIndex() == m.getEndIndex();
                ret.addElement(s.substring(p,zeroLength ? m.getStartIndex()+1 : m.getStartIndex()));
                p = zeroLength ? p + 1 : m.getEndIndex();
                if(!zeroLength) {
                    for(int i=1;i<=re.getNumSubs();i++) {
                        ret.addElement(m.toString(i));
                        if(ret.length() == limit) break OUTER;
                    }
                }
            } else {
                int x = s.indexOf(sep,p);
                if(x == -1) break OUTER;
                ret.addElement(s.substring(p,x));
                p = x + sep.length();
            }
            if(ret.length() == limit) break;
        }
        if(p < s.length() && ret.length() != limit)
            ret.addElement(s.substring(p));
        return ret;
    }
    
    public static GnuRegexp.RE newRE(String pattern, int flags) throws JSExn {
        try {
            return new GnuRegexp.RE(pattern,flags,GnuRegexp.RESyntax.RE_SYNTAX_PERL5);
        } catch(GnuRegexp.REException e) {
            throw new JSExn(e.toString());
        }
    }
    
    public String typeName() { return "regexp"; }
}
